/*
 * C6Run_mem.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Memory allocation routines for C6Run buffers for remote procedures   *
************************************************************************/

/*  ----------------------------------- Headers for User Code       */
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/*  ----------------------------------- DSP/BIOS Headers            */
#include "C6Runcfg.h"
#include <bcache.h>

/*  ----------------------------------- Application Headers         */
#include "debug_common.h"

/*  ----------------------------------- This module's header        */
#include "C6Run_mem.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/

#define C6RUN_MEM_MALLOC_MAGIC        (0x43365231)
#define C6RUN_MEM_MEMALIGN_MAGIC      (0x43365232)

#define SET_MALLOC_MAGIC(P)           (C6RUN_MEM_MALLOC_MAGIC ^ ((uint32_t)P))
#define IS_MALLOC_MAGIC(M,P)          ((((uint32_t)M) ^ ((uint32_t)P)) == C6RUN_MEM_MALLOC_MAGIC)
#define SET_MEMALIGN_MAGIC(P)         (C6RUN_MEM_MEMALIGN_MAGIC ^ ((uint32_t)P))
#define IS_MEMALIGN_MAGIC(M,P)        ((((uint32_t)M) ^ ((uint32_t)P)) == C6RUN_MEM_MEMALIGN_MAGIC)

#define MAX_SIZE_T                    (~(size_t)0)

#define NUM_TABLE_BINS_BITS           (12)
#define NUM_TABLE_BINS                (0x1 << NUM_TABLE_BINS_BITS)

#define MIN_ALIGNMENT_BITS            (3)
#define MIN_ALIGNMENT                 (0x1 << MIN_ALIGNMENT_BITS)

#define BUF2PTR(B) ((void *)((C6RUN_MEM_Buffer *)B + 1))
#define PTR2BUF(P) ((pC6RUN_MEM_Buffer)((C6RUN_MEM_Buffer *)P - 1))

// FIXME - implementing locking if we ever use this allocation code
#define C6RUN_MEM_MALLOC_LOCK() 
#define C6RUN_MEM_MALLOC_UNLOCK() 


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/

static void LOCAL_insertInBufferTable(void *ptr);
static void LOCAL_removeFromBufferTable(void *ptr);


/************************************************************
* Local Variable Definitions                                *
************************************************************/


/************************************************************
* Global Variable Definitions                               *
************************************************************/

volatile bool  C6RUN_MEM_initialized    = false;
// FIXME - these need to be set correctly (maybe by data passed from GPP)
void          *C6RUN_MEM_virtBaseAddr   = NULL;
void          *C6RUN_MEM_physBaseAddr   = NULL;

pC6RUN_MEM_Buffer C6RUN_MEM_bufferTable[NUM_TABLE_BINS];


/************************************************************
* Global Function Definitions                               *
************************************************************/

// Memory subsystem initialization routines
int C6RUN_MEM_init  ( void )
{
  int32_t status = 0;
  int i;

  if (C6RUN_MEM_initialized)
  {
    return status;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_MEM_init()\n");
 
  // Init the C6RUN_MEM_bufferTable to NULL values
  
  for (i=0; i<NUM_TABLE_BINS; i++)
  {
    C6RUN_MEM_bufferTable[i] = NULL;
  }

  C6RUN_MEM_initialized = true;
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_init()\n");
  
  return status;
}

void C6RUN_MEM_exit ( void )
{
  int i;
  
  if (!C6RUN_MEM_initialized)
  {
    return;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_MEM_exit()\n");

  // Check for orphaned memory(leaks) and print it out
  for (i=0; i<NUM_TABLE_BINS; i++)
  {
    pC6RUN_MEM_Buffer pBuffer = C6RUN_MEM_bufferTable[i];
    if (pBuffer != NULL)
    {
      do
      {  
        printf("Orphaned memory at bin %d, 0x%08X.\n",i,(uint32_t)BUF2PTR(pBuffer));
        pBuffer = pBuffer->next;
      } while (pBuffer != NULL);
    }
  }

  C6RUN_MEM_initialized = false;
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_exit()\n");
}

// C6Run memory buffer management routines
void *C6RUN_MEM_malloc(size_t size)
{
  size_t totalSize;
  void *tmpPtr;
  
  VERBOSE_PRINT1("C6RUN_MEM_malloc(%d)\n",size);
  
  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }  

  // Calculate total size
  totalSize = (size + sizeof(C6RUN_MEM_Buffer));
  
  // Call dlmalloc
  tmpPtr = malloc(totalSize);
  
  if (tmpPtr)
  {
    pC6RUN_MEM_Buffer pBuffer;
    
    C6RUN_MEM_MALLOC_LOCK();
    
    pBuffer = (pC6RUN_MEM_Buffer) tmpPtr;

    pBuffer->next = NULL;
    tmpPtr = BUF2PTR(tmpPtr);
    pBuffer->magic = SET_MALLOC_MAGIC(tmpPtr);
    
    // Insert the pBuffer into the buffer table
    LOCAL_insertInBufferTable(tmpPtr);
    
    C6RUN_MEM_MALLOC_UNLOCK();
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_malloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_realloc(void *oldPtr,size_t size)
{
  size_t totalSize;
  void *tmpPtr;
  pC6RUN_MEM_Buffer pBuffer;
  
  VERBOSE_PRINT2("C6RUN_MEM_realloc(0x%08X,%d)\n",(uint32_t)oldPtr,size);
  
  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  if (oldPtr == NULL)
  {
    return C6RUN_MEM_malloc(size);
  }
  
  // Calculate total size
  totalSize = (size + sizeof(C6RUN_MEM_Buffer));
  pBuffer = PTR2BUF(oldPtr);
  
  if (pBuffer && ! IS_MALLOC_MAGIC(pBuffer->magic,oldPtr))
  {
    // This could be a memaligned block
    if ( IS_MEMALIGN_MAGIC((((uint32_t *)oldPtr)[-1]),oldPtr) )
    {
      oldPtr = ((void**)oldPtr)[-2];
      pBuffer =  PTR2BUF(oldPtr);
    }  
    return NULL;
  }
  
  if (IS_MALLOC_MAGIC(pBuffer->magic,oldPtr))
  {
    tmpPtr = realloc((void *)pBuffer,totalSize);
  }
  else
  {
    tmpPtr = NULL;
  }
  
  if (tmpPtr)
  {
    pBuffer = (pC6RUN_MEM_Buffer) tmpPtr;
    tmpPtr = BUF2PTR(pBuffer);
    // If the new pointer is not the same as the old one, we have
    // some updates we need to do in the buffer table
    if (tmpPtr != oldPtr)
    {
      C6RUN_MEM_MALLOC_LOCK(); 
    
      // Remove old pointer from table
      LOCAL_removeFromBufferTable(oldPtr);
      
      pBuffer->next = NULL;
      pBuffer->magic = SET_MALLOC_MAGIC(tmpPtr);
      
      // Add new pointer/buffer to table
      LOCAL_insertInBufferTable(tmpPtr);
      
      C6RUN_MEM_MALLOC_UNLOCK();
    }
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_realloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_memalign(size_t alignment, size_t size)
{
  size_t totalSize;
  void *tmpPtr;
  
  VERBOSE_PRINT2("C6RUN_MEM_memalign(%d,%d)\n",alignment,size);

  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  // We can just use malloc
  if (alignment <= MIN_ALIGNMENT)
  {
    return C6RUN_MEM_malloc(size);
  }
  
  // Check for non power-of-two alignment
  if ((alignment & (alignment - 1)))
  {
    return NULL;
  }
  
  // Aligment is at least MIN_ALIGNMENT<<1 bytes
  totalSize = (alignment-MIN_ALIGNMENT) + size;
  tmpPtr = C6RUN_MEM_malloc(totalSize);
  if (tmpPtr != NULL)
  {
    intptr_t ptr = (intptr_t)tmpPtr;
    if ((ptr % alignment) == 0)
    {
      return tmpPtr;
    }

    // Align the pointer
    ptr += ((-ptr) % alignment);
    
    // There is always enough space for the base pointer and the memalign magic word
    ((uint32_t*)ptr)[-1] = SET_MEMALIGN_MAGIC(ptr);
    ((void**)ptr)[-2] = tmpPtr;

    tmpPtr = (void*)ptr;
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_memalign returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_calloc(size_t nelem, size_t elsize)
{
  size_t totalSize;
  void*  tmpPtr;
  
  VERBOSE_PRINT2("C6RUN_MEM_calloc(%d,%d)\n",nelem,elsize);
  
  totalSize = nelem * elsize;
  tmpPtr  = C6RUN_MEM_malloc(totalSize);
  if (tmpPtr != NULL)
  {
    memset(tmpPtr, 0x00, totalSize);
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_calloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void C6RUN_MEM_free(void* ptr)
{
  VERBOSE_PRINT1("C6RUN_MEM_free(0x%08X)\n",(uint32_t)ptr);

  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  if (ptr != NULL)
  {
    pC6RUN_MEM_Buffer pBuffer;
    C6RUN_MEM_MALLOC_LOCK(); 
    pBuffer = PTR2BUF(ptr);
  
    if (! IS_MALLOC_MAGIC(pBuffer->magic,ptr))
    {
      // This could be a memaligned block
      if ( IS_MEMALIGN_MAGIC((((uint32_t *)ptr)[-1]),ptr) )
      {
        ptr = ((void**)ptr)[-2];
        pBuffer = PTR2BUF(ptr);
      }
    }
  
    if (IS_MALLOC_MAGIC(pBuffer->magic,ptr))
    {
      LOCAL_removeFromBufferTable(ptr);
      free((void *)pBuffer);
    }

    C6RUN_MEM_MALLOC_UNLOCK(); 
  }
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_free.\n");
}


// Memory cache management routines
void C6RUN_MEM_globalInv()
{
  BCACHE_wbInvAll();
}

void C6RUN_MEM_globalWb()
{
  BCACHE_wbAll();
}

void C6RUN_MEM_inv(void *ptr,size_t size)
{
  BCACHE_inv(ptr, size, true);
}

void C6RUN_MEM_wb(void *ptr,size_t size)
{
  BCACHE_wb(ptr, size, true);
}


// Memory helper routines
void *C6RUN_MEM_lookupBuffer(void *ptr, size_t *buffSize)
{
  uint32_t binNum = (((uint32_t) ptr) >> MIN_ALIGNMENT_BITS) & (NUM_TABLE_BINS - 1);
  pC6RUN_MEM_Buffer pTmpBuffer = C6RUN_MEM_bufferTable[binNum];
  uint32_t base, size;
  
  // We check, but this should never happen
  if (pTmpBuffer)
  {
    // Search through buffer table bin for a match. A match is the
    // input pointer inside the range of an allocated buffer
    base = (uint32_t) BUF2PTR(pTmpBuffer);
    // Get size from dlmalloc header data (chunk size minus dlmalloc 
    // header size minus C6RUN_MEM_Buffer size)
    size = (uint32_t) (((size_t*)pTmpBuffer)[-1] & 0xFFFFFFFC) - (2 * sizeof(size_t)) - sizeof(C6RUN_MEM_Buffer);
    
    if ((base >= ((uint32_t) ptr)) && (((uint32_t) ptr) < (base + size)))
    {
      *buffSize = size;
      return ((void *)base);
    }
    else
    {
      while (pTmpBuffer->next != NULL)
      {
        base = (uint32_t) BUF2PTR(pTmpBuffer->next);
        size = (uint32_t) (((size_t*)pTmpBuffer->next)[-1] & 0xFFFFFFFC) - (2 * sizeof(size_t)) - sizeof(C6RUN_MEM_Buffer);
        if ((base >= ((uint32_t) ptr)) && (((uint32_t) ptr) < (base + size)))
        {
          *buffSize = size;
          return ((void *)base);
        }
        else
        {
          pTmpBuffer = pTmpBuffer->next;
        }
      }
    }
  }
  
  *buffSize = 0;
  return NULL;
}

void *C6RUN_MEM_virt2Phys(void *virtPtr)
{
  return (void *) ((((char *)virtPtr) - ((uint32_t)C6RUN_MEM_virtBaseAddr)) + ((uint32_t)C6RUN_MEM_physBaseAddr));
}

void *C6RUN_MEM_phys2Virt(void *physPtr)
{
  return (void *) ((((char *)physPtr) - ((uint32_t)C6RUN_MEM_physBaseAddr)) + ((uint32_t)C6RUN_MEM_virtBaseAddr));
}


/***********************************************************
* Local Function Definitions                               *
***********************************************************/

static void LOCAL_insertInBufferTable(void *ptr)
{
  uint32_t binNum = (((uint32_t) ptr) >> MIN_ALIGNMENT_BITS) & (NUM_TABLE_BINS - 1);
  pC6RUN_MEM_Buffer pTmpBuffer = C6RUN_MEM_bufferTable[binNum];  
  pC6RUN_MEM_Buffer pBuffer = PTR2BUF(ptr);
  
  VERBOSE_PRINT2("Insert - Bin: %d,Pointer: 0x%08X\n",binNum,(uint32_t)ptr);
  
  if (pTmpBuffer == NULL)
  {
    C6RUN_MEM_bufferTable[binNum] = pBuffer;
  }
  else
  {
    while (pTmpBuffer->next != NULL)
    {
      pTmpBuffer = pTmpBuffer->next;
      VERBOSE_PRINT1("         Buffer: 0x%08X\n",(uint32_t)pTmpBuffer);
    }
    pTmpBuffer->next = pBuffer;
  }
}

static void LOCAL_removeFromBufferTable(void *ptr)
{
  uint32_t binNum = (((uint32_t) ptr) >> MIN_ALIGNMENT_BITS) & (NUM_TABLE_BINS - 1);
  pC6RUN_MEM_Buffer pTmpBuffer = C6RUN_MEM_bufferTable[binNum];
  pC6RUN_MEM_Buffer pPtrBuffer = PTR2BUF(ptr);
  
  VERBOSE_PRINT2("Remove - Bin: %d,Pointer: 0x%08X\n",binNum,(uint32_t)ptr);

  // We check, but this should never happen
  if (!pTmpBuffer)
  {
    return;
  }
  
  if ( pPtrBuffer == pTmpBuffer)
  {
    C6RUN_MEM_bufferTable[binNum] = pTmpBuffer->next;
    return;
  }
  else
  {
    while (pTmpBuffer->next != NULL)
    {
      if (pPtrBuffer == pTmpBuffer->next)
      {
        pTmpBuffer->next = pTmpBuffer->next->next;
        return;
      }
      pTmpBuffer = pTmpBuffer->next;
      VERBOSE_PRINT1("         Buffer: 0x%08X\n",(uint32_t)pTmpBuffer);
    }
  }
}


/***********************************************************
* End file                                                 *
***********************************************************/
